from solaredge2mqtt.models.base import Component, EnumModel
from solaredge2mqtt.models.modbus import SunSpecBattery, SunSpecInverter, SunSpecMeter
from solaredge2mqtt.models.monitoring import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.models.powerflow import Powerflow
from solaredge2mqtt.models.wallbox import WallboxAPI



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the gmqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""
import asyncio
import signal
import datetime as dt

from typing import Optional

from scheduler.asyncio import Scheduler

from solaredge2mqtt.api import MonitoringSite
from solaredge2mqtt.logging import initialize_logging, logger
from solaredge2mqtt.influxdb import InfluxDB
from solaredge2mqtt.modbus import Modbus
from solaredge2mqtt.models import Powerflow
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.wallbox import WallboxClient
from solaredge2mqtt.settings import service_settings

STOP = asyncio.Event()

settings = service_settings()


def run():
    """Initializes and starts the SolarEdge2MQTT service."""
    initialize_logging(settings.logging_level)

    logger.info("Starting SolarEdge2MQTT service...")

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, ask_stop)
    loop.add_signal_handler(signal.SIGTERM, ask_stop)
    loop.run_until_complete(main())
    loop.close()


def ask_stop():
    """Stops the SolarEdge2MQTT service by setting the STOP event."""
    logger.info("Stopping SolarEdge2MQTT service...")
    STOP.set()


async def main():
    """
    Initializes the SolarEdge inverter and logs the connection details.
    This function is the main entry point for the SolarEdge2MQTT service.
    """

    modbus = Modbus(settings)

    if settings.is_wallbox_configured:
        wallbox = WallboxClient(settings)
    else:
        wallbox = None

    if settings.is_influxdb_configured:
        influxdb = InfluxDB(settings)
        influxdb.initialize_buckets()
        influxdb.initialize_task()
    else:
        influxdb = None

    mqtt = MQTTClient(settings)

    await mqtt.connect()

    loop = asyncio.get_running_loop()
    scheduler = Scheduler(loop=loop)
    scheduler.cyclic(
        dt.timedelta(seconds=settings.interval),
        modbus_and_wallbox_loop,
        args=[modbus, mqtt, wallbox, influxdb],
    )

    if settings.is_api_configured:
        monitoring = MonitoringSite(settings)
        monitoring.login()
        await energy_loop(monitoring, mqtt)
        scheduler.cyclic(
            dt.timedelta(seconds=300), energy_loop, args=[monitoring, mqtt]
        )

    logger.debug(scheduler)

    while not STOP.is_set():
        await asyncio.sleep(1)

    scheduler.delete_jobs()
    await mqtt.disconnect()


modbus_lock = asyncio.Lock()


async def modbus_and_wallbox_loop(
    modbus: Modbus,
    mqtt: MQTTClient,
    wallbox: Optional[WallboxClient] = None,
    influxdb: Optional[InfluxDB] = None,
):
    if modbus_lock.locked():
        logger.warning("Modbus is still locked, skipping this loop")
        return

    async with modbus_lock:
        results = await asyncio.gather(
            modbus.loop(),
            wallbox.loop() if settings.is_wallbox_configured else asyncio.sleep(0),
        )

        inverter_data, meters_data, batteries_data = results[0]

        if any(data is None for data in [inverter_data, meters_data, batteries_data]):
            logger.warning("Invalid modbus data, skipping this loop")
            return

        wallbox_data = results[1]
        evcharger = 0

        if wallbox_data is not None:
            evcharger = wallbox_data.power

        powerflow = Powerflow.calc(
            inverter_data, meters_data, batteries_data, evcharger
        )
        if not powerflow.consumer.is_valid:
            logger.warning("Invalid powerflow data: {powerflow}", powerflow=powerflow)

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, House {consumer.house} W, "
            + "Grid {grid.power} W, Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        mqtt.publish_inverter(inverter_data)
        mqtt.publish_meters(meters_data)
        mqtt.publish_batteries(batteries_data)
        mqtt.publish_powerflow(powerflow)

        if wallbox_data is not None:
            mqtt.publish_wallbox(wallbox_data)

        if influxdb is not None:
            influxdb.write_components(
                inverter_data, meters_data, batteries_data, wallbox_data
            )

            influxdb.write_powerflow(powerflow)
            influxdb.flush_loop()


energy_lock = asyncio.Lock()


async def energy_loop(monitoring: MonitoringSite, mqtt: MQTTClient):
    """Publishes the energy data from monitoring site to the MQTT broker."""
    if energy_lock.locked():
        logger.warning("Energy is still locked, skipping this loop")
        return

    async with energy_lock:
        modules = monitoring.get_module_energies()

        if modules is None:
            logger.warning("Invalid monitoring data, skipping this loop")
            return

        energy_total = 0
        count_modules = 0
        for module in modules:
            if module.energy is not None:
                count_modules += 1
                energy_total += module.energy

        logger.info(
            "Read from monitoring total energy: {energy_total} kWh from {count_modules} modules",
            energy_total=energy_total / 1000,
            count_modules=count_modules,
        )

        mqtt.publish_pv_energy_today(energy_total)
        mqtt.publish_module_energy(modules)



from typing import Dict

from gmqtt import Client, Message
from pydantic import BaseModel

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    LogicalModule,
    Powerflow,
    SunSpecBattery,
    SunSpecInverter,
    SunSpecMeter,
    WallboxAPI,
)
from solaredge2mqtt.settings import ServiceSettings


class MQTTClient:
    def __init__(self, settings: ServiceSettings):
        self.broker = settings.broker
        self.port = settings.port

        self.topic_prefix = settings.topic_prefix

        logger.info(
            "Using MQTT broker: {broker}:{port}",
            broker=settings.broker,
            port=settings.port,
        )

        self.connected = False

        will_message = Message(
            f"{settings.topic_prefix}/status",
            b"offline",
            qos=1,
            retain=True,
            will_delay_interval=10,
        )
        self.client = Client(settings.client_id, will_message=will_message)
        self.client.set_auth_credentials(settings.username, settings.password)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

    async def connect(self) -> None:
        await self.client.connect(self.broker, self.port)

    async def disconnect(self) -> None:
        self.client.publish(
            f"{self.topic_prefix}/status", "offline", qos=1, retain=True
        )
        await self.client.disconnect()

    def on_connect(self, client: Client, flags, rc, properties) -> None:
        # pylint: disable=unused-argument
        logger.info("Connected to MQTT broker")

        client.publish(f"{self.topic_prefix}/status", "online", qos=1, retain=True)
        self.connected = True

    def on_disconnect(self, client: Client, packet, exc=None) -> None:
        # pylint: disable=unused-argument
        logger.info("Disconnected from MQTT broker")
        self.connected = False

    def publish_inverter(self, inverter: SunSpecInverter) -> None:
        self._publish("modbus/inverter", inverter)

    def publish_meters(self, meters: Dict[str, SunSpecMeter]) -> None:
        for meter_key, meter in meters.items():
            self._publish(f"modbus/meter/{meter_key.lower()}", meter)

    def publish_batteries(self, batteries: Dict[str, SunSpecBattery]) -> None:
        for battery_key, battery in batteries.items():
            self._publish(f"modbus/battery/{battery_key.lower()}", battery)

    def publish_wallbox(self, wallbox: WallboxAPI) -> None:
        self._publish("rest/wallbox", wallbox)

    def publish_powerflow(self, powerflow: Powerflow) -> None:
        self._publish("powerflow", powerflow)

    def publish_pv_energy_today(self, energy: int) -> None:
        self._publish("api/monitoring/pv_energy_today", energy)

    def publish_module_energy(self, modules: list[LogicalModule]) -> None:
        for module in modules:
            self._publish(
                f"api/monitoring/module/{module.info.serialnumber}",
                module,
            )

    def _publish(self, topic: str, payload: str | int | float | BaseModel) -> None:
        if self.connected:
            if isinstance(payload, BaseModel):
                payload = payload.model_dump_json()
            self.client.publish(f"{self.topic_prefix}/{topic}", payload, qos=1)



from influxdb_client import (
    InfluxDBClient,
    Point,
    BucketRetentionRules,
    BucketsApi,
    TaskCreateRequest,
)

import pkg_resources

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import Component, Powerflow
from solaredge2mqtt.settings import ServiceSettings


class InfluxDB:
    def __init__(self, settings: ServiceSettings) -> None:
        self.prefix: str = settings.influxdb_prefix

        self.retention_raw: int = settings.influxdb_retention_raw
        self.retention_aggregated: int = settings.influxdb_retention_aggregated

        self.org: str = settings.influxdb_org

        self.client: InfluxDBClient = InfluxDBClient(
            url=f"{settings.influxdb_host}:{settings.influxdb_port}",
            token=settings.influxdb_token,
            org=settings.influxdb_org,
        )

        self.loop_points: list[Point] = []

        self.flux_cache: dict[str, str] = {}

    def initialize_buckets(self) -> None:
        buckets_api = self.client.buckets_api()
        self._create_or_update_bucket(buckets_api, self.bucket_raw, self.retention_raw)
        self._create_or_update_bucket(
            buckets_api, self.bucket_aggregated, self.retention_aggregated
        )

    def _create_or_update_bucket(
        self, buckets_api: BucketsApi, bucket_name: str, retention: int
    ) -> None:
        bucket = buckets_api.find_bucket_by_name(bucket_name)
        retention_rules = BucketRetentionRules(type="expire", every_seconds=retention)

        if bucket is None:
            logger.info(f"Creating bucket '{bucket_name}'")
            buckets_api.create_bucket(
                bucket_name=bucket_name, retention_rules=retention_rules
            )
        else:
            logger.info(f"Bucket '{bucket_name}' already exists.")
            if bucket.retention_rules[0].every_seconds != retention:
                logger.info(
                    f"Updating retention rules for bucket '{bucket_name}' to {retention} seconds."
                )
                bucket.retention_rules[0] = retention_rules
                buckets_api.update_bucket(bucket=bucket)

    @property
    def bucket_raw(self) -> str:
        return f"{self.prefix}_raw"

    @property
    def bucket_aggregated(self) -> str:
        return f"{self.prefix}"

    def initialize_task(self) -> None:
        tasks_api = self.client.tasks_api()
        tasks = tasks_api.find_tasks(name=self.task_name)
        if not tasks:
            flux = self._get_flux_query("aggregation").replace(
                "TASK_NAME", self.task_name
            )

            logger.info(f"Creating task '{self.task_name}'")
            logger.debug(flux)
            task_request = TaskCreateRequest(
                flux=flux, org_id=self.org, status="active"
            )
            tasks_api.create_task(task_create_request=task_request)
        else:
            logger.info(f"Task '{self.task_name}' already exists.")

    def _get_flux_query(self, query_name: str) -> str:
        if query_name not in self.flux_cache:
            flux = pkg_resources.resource_string(
                __name__, f"resources/{query_name}.flux"
            ).decode("utf-8")
            flux = flux.replace("BUCKET_RAW", self.bucket_raw).replace(
                "BUCKET_AGGREGATED", self.bucket_aggregated
            )
            self.flux_cache[query_name] = flux

        return self.flux_cache[query_name]

    @property
    def task_name(self) -> str:
        return f"{self.prefix}_aggregation"

    def write_components(
        self, *args: list[Component | dict[str, Component] | None]
    ) -> None:
        for component in args:
            if isinstance(component, Component):
                self.loop_points.append(self._create_component_point(component))
            elif isinstance(component, dict):
                for name, component in component.items():
                    self.loop_points.append(
                        self._create_component_point(component, {"name": name})
                    )

    def write_component(self, component: Component) -> None:
        self.loop_points.append(self._create_component_point(component))

    def _create_component_point(
        self, component: Component, additional_tags: dict[str, str] = None
    ) -> Point:
        point = Point("component")

        for key, value in component.influxdb_tags.items():
            point.tag(key, value)

        if additional_tags is not None:
            for key, value in additional_tags.items():
                point.tag(key, value)

        for key, value in component.influxdb_fields().items():
            point.field(key, value)

        return point

    def write_powerflow(self, powerflow: Powerflow) -> None:
        point = Point("powerflow")
        for key, value in powerflow.influxdb_fields().items():
            point.field(key, value)
        self.loop_points.append(point)

    def flush_loop(self) -> None:
        write_api = self.client.write_api()
        write_api.write(bucket=self.bucket_raw, record=self.loop_points)
        self.loop_points = []

    def energy_loop(self) -> None:
        pass



from __future__ import annotations

from typing import Dict, Optional

from solaredge2mqtt.models.base import InfluxDBModel
from solaredge2mqtt.models.modbus import SunSpecBattery, SunSpecInverter, SunSpecMeter


class InverterPowerflow(InfluxDBModel):
    power: int
    consumption: int
    production: int
    pv_production: int
    battery_production: int

    @staticmethod
    def calc(
        inverter_data: SunSpecInverter,
        battery: BatteryPowerflow,
    ) -> InverterPowerflow:
        power = int(inverter_data.ac.power.actual)

        if power >= 0:
            consumption = 0
            production = power
            if battery.discharge > 0:
                battery_factor = battery.discharge / inverter_data.dc.power
                battery_production = int(round(production * battery_factor))
                battery_production = min(battery_production, production)
                pv_production = production - battery_production
            else:
                battery_production = 0
                pv_production = production

        else:
            consumption = int(abs(power))
            production = 0
            pv_production = 0
            battery_production = 0

        return InverterPowerflow(
            power=power,
            consumption=consumption,
            production=production,
            pv_production=pv_production,
            battery_production=battery_production,
        )


class GridPowerflow(InfluxDBModel):
    power: int
    consumption: int
    delivery: int

    @staticmethod
    def calc(meters_data: Dict[str, SunSpecMeter]) -> GridPowerflow:
        grid = 0
        for meter in meters_data.values():
            if "Import" in meter.info.option and "Export" in meter.info.option:
                grid += meter.power.actual

        if grid >= 0:
            consumption = 0
            delivery = grid
        else:
            consumption = int(abs(grid))
            delivery = 0

        return GridPowerflow(power=grid, consumption=consumption, delivery=delivery)


class BatteryPowerflow(InfluxDBModel):
    power: int
    charge: int
    discharge: int

    @staticmethod
    def calc(batteries_data: Dict[str, SunSpecBattery]) -> BatteryPowerflow:
        batteries_power = 0
        for battery in batteries_data.values():
            batteries_power += battery.power

        if batteries_power >= 0:
            charge = batteries_power
            discharge = 0
        else:
            charge = 0
            discharge = abs(batteries_power)

        return BatteryPowerflow(
            power=batteries_power, charge=charge, discharge=discharge
        )


class ConsumerPowerflow(InfluxDBModel):
    house: int
    evcharger: int = 0
    inverter: int

    total: int

    used_pv_production: int
    used_battery_production: int

    @staticmethod
    def calc(
        inverter: InverterPowerflow, grid: GridPowerflow, evcharger: int
    ) -> ConsumerPowerflow:
        house = int(abs(grid.power - inverter.power))
        if evcharger < house:
            house -= evcharger
        else:
            # Happens when EV Charger starts up and meters are not yet updated
            evcharger = 0

        total = house + evcharger + inverter.consumption

        if inverter.pv_production > inverter.production - grid.delivery:
            pv_production = inverter.pv_production - grid.delivery
        else:
            pv_production = inverter.pv_production

        if inverter.battery_production > inverter.production - grid.delivery:
            battery_production = inverter.battery_production - grid.delivery
        else:
            battery_production = inverter.battery_production

        return ConsumerPowerflow(
            house=house,
            evcharger=evcharger,
            used_pv_production=pv_production,
            used_battery_production=battery_production,
            inverter=inverter.consumption,
            total=total,
        )

    def is_valid(self) -> bool:
        return self.total >= self.used_battery_production + self.used_pv_production


class Powerflow(InfluxDBModel):
    pv_production: int
    inverter: InverterPowerflow
    grid: GridPowerflow
    battery: BatteryPowerflow
    consumer: ConsumerPowerflow

    @staticmethod
    def calc(
        inverter_data: SunSpecInverter,
        meters_data: Dict[str, SunSpecMeter],
        batteries_data: Dict[str, SunSpecBattery],
        evcharger: Optional[int] = 0,
    ) -> Powerflow:
        grid = GridPowerflow.calc(meters_data)
        battery = BatteryPowerflow.calc(batteries_data)

        if inverter_data.ac.power.actual > 0:
            pv_production = int(inverter_data.dc.power + battery.power)
            if pv_production < 0:
                pv_production = 0
        else:
            pv_production = 0

        inverter = InverterPowerflow.calc(inverter_data, battery)

        consumer = ConsumerPowerflow.calc(inverter, grid, evcharger)

        return Powerflow(
            pv_production=pv_production,
            inverter=inverter,
            grid=grid,
            battery=battery,
            consumer=consumer,
        )



from __future__ import annotations
from typing import Any, Dict, Optional, ClassVar

from enum import Enum

from pydantic import BaseModel, model_serializer


class EnumModel(Enum):
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __str__(self) -> str:
        return str(self._value_)

    def __repr__(self) -> str:
        return str(self._value_)

    def __eq__(self, obj) -> bool:
        return isinstance(obj, self.__class__) and self.value == obj.value

    def __hash__(self):
        return hash(self.value)

    @classmethod
    def from_string(cls, value: str):
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"No enum value {value} found.")

    @model_serializer
    def serialize(self):
        return self.value


class InfluxDBModel(BaseModel):
    def influxdb_fields(self, prefix: Optional[str] = None) -> Dict[str, Any]:
        fields = {}
        for attr, value in self.__dict__.items():
            if attr == "info":
                continue
            if isinstance(value, InfluxDBModel):
                attr_name = attr
                if prefix is not None:
                    attr_name = f"{prefix}_{attr_name}"
                fields = {**fields, **value.influxdb_fields(attr_name)}
            else:
                if prefix is not None:
                    attr = f"{prefix}_{attr}"
                fields[attr] = float(value) if isinstance(value, int) else value
        return fields


class ComponentValueGroup(InfluxDBModel):
    @staticmethod
    def scale_value(
        data: Dict[str, str | int],
        value_key: str,
        scale_key: Optional[str] = None,
        digits: Optional[int] = 2,
    ) -> float:
        if scale_key is None:
            scale_key = f"{value_key}_scale"

        value = int(data[value_key])
        scale = int(data[scale_key])

        return round(value * 10**scale, digits)

    # @staticmethod
    # def is_three_phase(data: Dict[str, Any]) -> bool:
    #     return data["c_sunspec_did"] in [
    #         sunspecDID.THREE_PHASE_INVERTER.value,
    #         sunspecDID.WYE_THREE_PHASE_METER.value,
    #         sunspecDID.DELTA_THREE_PHASE_METER.value,
    #     ]


class Component(ComponentValueGroup):
    COMPONENT: ClassVar[str] = "unknown"
    SOURCE: ClassVar[str] = "unknown"

    @property
    def influxdb_tags(self) -> Dict[str, str]:
        return {
            "component": self.COMPONENT,
            "source": self.SOURCE,
        }
